( FLENG - uxn/varvara runtime system
;
; register usage:
;
;   A: <TOS>
;   G: GOAL
;   N: NEWGOAL
;   P0-P3: PARAM[<I>]
;   E0-E<n>: ENV[<I>]
;
; encoding:
;
;   string:             aaaaaaaa aaaaaaa1
;   immediate integer:  0snnnnnn nnnnnnn1
;   cell:               1aaaaaaa aaaaaaa0
;
;   cell = [TAG, CAR, CDR]
;
;   tag:    _TLIVAAA cccccccc
;
;   T = 1: tuple ("AAA" is arity / length - 1)
;   V = 1: var
;   I = 1: integer
;   L = 1: list
;
;   List = [Car, Cdr]
;   Tuple = [Head, Arglist]
;   Integer = [Magnitude, _]
;   Var = [Val|Var, S-list]
;
;   S-List = [[Goal, [Var', Addr], ...], ...]
;
;   Var' = marked var (lowest bit set)
;
; type-ordering: int < string < list < tuple )

%IMMEDIATE_BIT { #0001 }
%LOW_BIT { #8000 }
%SIGN_BIT { #4000 }

%VAR_BIT { #0800 }
%INT_BIT { #1000 }
%LIST_BIT { #2000 }
%TUPLE_BIT { #4000 }

%TIMESLICE { #0a }

%TAG_MASK { #ff00 }
%COUNT_MASK { #00ff }
%ARITY_MASK { #0700 }
%BITS_MASK { #f800 }

%MAX_ENV { #20 }
%HEAP { #8000 }
%MAX_CELLS { #1500 }
%MAX_GOALS { #0400 }

%goalbuf { #5700 }  ( (HEAP - (MAX_GOALS + 1) * 5 * 2) & $ff00 )
%sstacktop { goalbuf }
%sstack { #5600 }   ( goalbuf - 256 )
%queuetop { sstack }
%queue { #4e00 }    ( (queuetop - 2 * MAX_GOALS) & $ff00 )

%emit { #18 DEO }
%debug { #18 DEO #010e DEO }

( zeropage )
|00
@UNUSED $2
@A $2
@X0 $2
@X1 $2
@AP $2
@GOAL $2
@NGOAL $2
@PARAM0 $2
@PARAM1 $2
@PARAM2 $2
@PARAM3 $2
@qstart $2
@qend $2
@sstackp $2
@goals $2
@cells $2
@suspended $2
@ticks $2
@idle $2
@epoch $1
@tailcalls $1
@timeslice $1
@hooked $1
@freegoals $2
@freelist $2
@vectors $40    ( [count2, var2, ...] )
@ENV $40

%not2 { #0000 EQU2 }
%immediate { IMMEDIATE_BIT AND2 NIP }
%lowptr { LOW_BIT AND2 #0000 EQU2 }
%reset_stack { #0002 DEO #0003 DEO }
%CAR { #0002 ADD2 LDA2 }
%CDR { #0004 ADD2 LDA2 }

|100

    ( entry point )
    ;init JMP2

    ( extract integer value )
    ( x2 -- int2 )
@fl_getint
    DUP2 immediate ,&imm JCN
    ( not immediate )
    #0002 ADD2
    LDA2
    JMP2r
&imm ( x2 -- int2 ) 
    DUP2 SIGN_BIT AND2 POP ,&signed JCN
    ( unsigned )
    #7fff AND2
    #01 SFT2
    JMP2r
&signed ( x2 -- int2 )
    #01 SFT2
    #e000 ORA2
    JMP2r

    ( create integer )
    ( int2 -- X2 )
@fl_mkint
    OVR DUP #10 SFT EOR #80 AND #00 EQU ,&imm JCN
    #0000 SWP2 INT_BIT
    ,fl_cons JMP
&imm ( int2 -- X2 )
    #10 SFT2
    #8001 ORA2
    JMP2r
    
    ( allocate cell )
    ( cdr2 car2 tag2 -- X2 )
@fl_cons
    .freelist LDZ2 DUP2 ORA ,&l1 JCN
    ;ENOCELLS ;error JMP2
&l1 ( cdr2 car2 tag2 freelist2 )
    DUP2 LDA2 .freelist STZ2
    ( cdr2 car2 tag2 freelist2 )
    STA2k
    NIP2 #0002 ADD2 
    SWP2 ;fl_addref JSR2
    SWP2 STA2k
    NIP2 #0002 ADD2
    SWP2 ;fl_addref JSR2
    SWP2 STA2k
    NIP2
    #0004 SUB2
    .cells LDZ2k INC2 ROT STZ2
    JMP2r

    ( create variable )
    ( _ -- var2 )
@op_mkvar
    POP2
    ,fl_mkvar JSR
    ;next JMP2

    ( -- var2 )
@fl_mkvar
    ;fl_nil #0000 VAR_BIT ,fl_cons JSR
    DUP2 #0002 ADD2 STA2k POP2
    JMP2r

    ( runtime error )
    ( errmsg2 -- )
@error
    ;&errhdr ,writestr JSR
    ,writestr JSR
    ;&errinfo1 ,writestr JSR
    .GOAL LDZ2 LDA2 ;writehex JSR2
    #3a emit
    .AP LDZ2 ;writehex JSR2
    ;&errinfo2 ,writestr JSR
    #010e DEO
    ;exit JMP2
&errhdr 0a "ERROR: 20 00
&errinfo1 20 "[ 00
&errinfo2 "] 0a 00

    ( write string to current output device )
    ( string2 -- )
@writestr
    LDAk DUP #00 EQU ,&done JCN
    emit
    INC2
    ;writestr JMP2
&done
    POP POP2
    JMP2r

    ( write P0 (string) and P1 (hex) )
@op_dbg
    #010e DEO
    .PARAM0 LDZ2 ;fl_check_atom JSR2
    INC2 ;writestr JSR2
    .PARAM1 LDZ2 
    ;fl_deref JSR2
    ;writehex JSR2
    #0a18 DEO
    ;next JMP2

    ( init vectors )
@fl_clear_vectors
    .vectors
&loop
    .ENV EQUk ,&done JCN
    POP #0000 ROT STZ2k
    NIP NIP INC INC
    ,&loop JMP
&done
    POP2
    JMP2r

    ( initialize runtime system )
@fl_initsys
    ( init vars )
    #0000 DUP2 .cells STZ2
    DUP2 .goals STZ2
    .suspended STZ2
    TIMESLICE .timeslice STZ
    #00 .tailcalls STZ
    #00 .hooked STZ
    #02 .epoch STZ
    ;fl_clear_vectors JSR2
    ( init sstack )
    sstack ;sstackp STA2
    ( init goals )
    goalbuf MAX_GOALS #0a ,linkblocks JSR
    .freegoals STZ2
    ( init heap )
    HEAP MAX_CELLS #06 ,linkblocks JSR
    .freelist STZ2
    ( init queue )
    queue DUP2 .qstart STZ2 .qend STZ2
    ;fl_mkvar JSR2 ;fl_addref JSR2
    .idle STZ2
    JMP2r

    ( link sequence of blocks together with ptr in first word )
    ( startaddr2 count2 blocksize -- ptr2 )
    ( modifies: X0, X1 )
@linkblocks
    #00 SWP .X0 STZ2 ( blocksize )
    #0000 .X1 STZ2 ( prev )
&loop ( addr2 count2 -- )
    SWP2 .X1 LDZ2 OVR2 STA2
    DUP2 .X1 STZ2
    .X0 LDZ2 ADD2 SWP2
    #0001 SUB2
    DUP2 ORA ,&loop JCN
    POP2 POP2
    .X1 LDZ2
    JMP2r

    ( increase refcount )
    ( X2 -- X2 )
@fl_addref
    DUP2 immediate ,fl_addref0/done JCN
    DUP2 lowptr ,fl_addref0/done JCN
@fl_addref0 ( block2 -- block2 )
    DUP2 INC2 LDA #ff EQU ,&done JCN
    DUP2 LDA2 INC2 OVR2 STA2
&done
    JMP2r

    ( decrease refcount )
    ( val2 -- )
@fl_unref
    DUP2 immediate ,&done JCN
    DUP2 lowptr ,&done JCN
    DUP2 INC2 LDA #ff EQU ,&done JCN
    DUP2 LDA2 #0001 SUB2 DUP
    #00 EQU ,&release JCN
    OVR2 STA2
&done
    POP2
    JMP2r
&release ( val2 tag2 -- )
    OVR2
    .freelist LDZ2 SWP2 STA2
    OVR2 .freelist STZ2
    .cells LDZ2 #0001 SUB2 .cells STZ2
    DUP2 INT_BIT AND2 #0000 EQU2 ,&rel2 JCN
    ( integer )
    POP2 POP2 JMP2r
&rel2
    VAR_BIT AND2 #0000 EQU2 ,&relb JCN
    ( variable )
    DUP2 INC2 INC2 LDA2 OVR2 NEQ2 ,&relb JCN
    ( unbound )
    #0004 ADD2 LDA2 ,fl_unref JMP
&relb ( block )
    DUP2 #0004 ADD2 LDA2
    SWP2
    INC2 INC2 LDA2
    ;fl_unref JSR2
    ;fl_unref JMP2

    ( dereference value )
    ( X -- X' )
@fl_deref
    DUP2 immediate ,&done JCN
    DUP2 LDA2 VAR_BIT AND2 #0000 EQU2 ,&done JCN
    DUP2 #0002 ADD2 LDA2 EQU2k ,&unb JCN
    ( var2 val2 )
    NIP2 
    ,fl_deref JMP
&unb
    POP2
&done
    JMP2r

    ( force variables )
    ( X2 -- X2' F != 0 when suspending )
@fl_force
    DUP2 immediate ,false JCN
    DUP2 LDA2 VAR_BIT AND2 #0000 EQU2 ,false JCN
    DUP2 #0002 ADD2 LDA2 EQU2k ,&spush JCN
    NIP2 
    ,fl_force JMP
&spush
    POP2
@sspush ( var2 -- 01 )
    .sstackp LDZ2 STA2k
    #0002 ADD2
    .GOAL LDZ2 LDA2 OVR2 STA2
    #0002 ADD2
    DUP2 .sstackp STZ2
    sstacktop NEQ2 ,true JCN
    ;ESSFULL ;error JMP2

@true 
    #01 JMP2r

@false
    #00 JMP2r

    ( allocate goal )
    ( -> N )
@fl_opengoal
    .freegoals LDZ2
    DUP2 ORA ,&open JCN
    ;ENOGOALS ;error JMP2
&open ( freegoals )
    .NGOAL STZ2k POP
    LDA2 .freegoals STZ2
    .goals LDZ2 INC2 .goals STZ2
    ( fall through ... )

    ( enqueue goal )
    ( N=goal )
@enqueue
    .NGOAL LDZ2 .qend LDZ2 STA2k
    NIP2 #0002 ADD2 .qend STZ2k
    POP DUP2 .qstart LDZ2 NEQ2 ,&wrap JCN
    ;EQUEUEFULL ;error JMP2
&wrap ( qend2 )
    queuetop NEQ2 ,&done JCN
    queue .qend STZ2
&done
    JMP2r

    ( dequeue goal )
    ( -> G )
@dequeue
    .qstart LDZ2 LDA2k .GOAL STZ2
    #0002 ADD2 queuetop NEQ2k ,&nowrap JCN
    POP2 POP2 queue DUP2
&nowrap
    POP2 .qstart STZ2
    JMP2r

    ( prepare for tail call )
    ( <AP>=addr -> N )
@op_tailgoal
    .tailcalls LDZ .timeslice LDZ NEQ ,&tail JCN
    ;fl_opengoal JSR2
    ;fetch JSR2
&setaddr ( addr2 )
    .NGOAL LDZ2 STA2
    ;next JMP2
&tail
    ;fetch JSR2
    .GOAL LDZ2 .NGOAL STZ2
    ,&setaddr JMP

    ( invoke predicate if timeslice is not exhausted )
    ( G=N )
@op_tailjump
    .tailcalls LDZ .timeslice LDZ NEQ ,&jump JCN
    #00 .tailcalls STZ
    ;op_dropgoal JMP2
&jump
    .tailcalls LDZ INC .tailcalls STZ
    ( fall through ... )

    ( invoke predicate )
    ( N )
@op_invoke_jump
    reset_stack
    #0000 ( dummy A )
    .NGOAL LDZ2 LDA2 .AP STZ2
    ;next JMP2

    ( prepare invoking predicate )
    ( G -> N, A -> addr(N) )
@op_invoke_goal
    DUP2
    .GOAL LDZ2 .NGOAL STZ2
    #7ffe AND2 .NGOAL LDZ2 STA2
    ;next JMP2

    ( N, A=list )
@op_puttailarg
    DUP2 #0004 ADD2 LDA2
    immediate ,&imm JCN
&nimm
    ;fl_addref JSR2
    .NGOAL LDZ2 #0008 ADD2 STA2
    ;next JMP2
&imm
    #0002 ADD2 LDA2
    ,&nimm JMP

    ( drop goal & yield )
    ( G )
@op_dropgoal
    .freegoals LDZ2 .GOAL LDZ2 STA2k
    NIP2
    .freegoals STZ2
    .goals LDZ2k INC2 ROT STZ2
    ,op_yield JMP

    ( does variable have active suspension? )
    ( var2 -- f )
@active
    #0004 ADD2 LDA2
    DUP2 ;fl_nil EQU2 ,&no JCN
    DUP2 #0002 ADD2 LDA2
    ( check goalptr from CAR for being 0 )
    #0002 ADD2 LDA2 #0000 EQU2 ,active JCN
    POP2
    #01
    JMP2r
&no
    POP2
    #00
    JMP2r

    ( yield to next process )
@op_yield
    ( #2a debug )
    .ticks LDZ2 INC2 .ticks STZ2
    .qstart LDZ2 .qend LDZ2 NEQ2 ,&alive JCN
    ( queue empty, check idle var for having suspensions )
    .idle LDZ2
    ,active JSR #00 EQU ,&check JCN
    ;fl_nil .idle LDZ2 ;fl_assign JSR2
    .idle LDZ2 ;fl_unref JSR2
    .epoch LDZk INC SWP STZ
    ;fl_mkvar JSR2 
    ;fl_addref0 JSR2 
    .idle STZ2
&check
    ( check queue again )
    .qstart LDZ2 .qend LDZ2 NEQ2 ,&alive JCN
    .hooked LDZ #00 EQU ,terminate JCN
    BRK
&alive
    ;dequeue JSR2
    ( enter goal code )
    reset_stack
    #0000 ( dummy A )
    .GOAL LDZ2 LDA2 .AP STZ2
    ;next JMP2

    ( add goal to pool )
    ( addr2 -- )
@fl_initgoal
    ;fl_opengoal JSR2
    .NGOAL LDZ2 STA2
    JMP2r

@terminate
    .suspended LDZ2 #0000 EQU2 ,&term1 JCN
    ;EDEADLOCK ;error JMP2
&term1
    .idle LDZ2 ;fl_unref JSR2
    ( unref vector streams )
    .vectors
&loop
    .ENV EQUk ,&done JCN
    POP #02 ADD LDZ2k 
    DUP2 #0000 EQU2 ,&next JCN 
    ;fl_unref JSR2
    #0000
&next
    POP2
    #02 ADD
    ,&loop JMP
&done
    .cells LDZ2 ORA ,eleak JCN
@exit
    ;fl_clear_vectors JSR2
    #010f DEO
    BRK
@eleak
    ;ELEAK ;error JMP2

    ( assign value to variable )
    ( val2 var2 -- )
@fl_assign
    OVR2 OVR2 #0002 ADD2 LDA2 NEQ2 ,&assign JCN
    POP2 POP2 
    JMP2r
&assign ( val2 var2 )
    DUP2 #0002 ADD2 LDA2 EQU2k ,&unbound JCN
    ;EBOUND ;error JMP2
&unbound ( val2 var2 var2 )
    ROT2 ;fl_addref JSR2
    SWP2       ( var2 val2 var2 )
    #0002 ADD2 STA2
    ;resume JMP2

    ( suspend on variables on sstack & yield )
    ( G )
@op_suspend
    ;fl_nil ,&slst STR2
    ,&build_slst JSR
    ,&slst LDR2
    .GOAL LDZ2 
    LIST_BIT
    ;fl_cons JSR2       ( slst = cons(goal, slst) )
    ,&slst STR2
    ,&attach JSR
    sstack .sstackp STZ2
    .suspended LDZ2 INC2 .suspended STZ2
    ;op_yield JMP2
&slst $2
&build_slst
    ;for_sstack JSR2
    ( build slst )
    ( var2 addr2 -- )
    SWP2
    #0001 ORA2
    LIST_BIT
    ;fl_cons JSR2       ( x = cons(var', addr) )
    ,&slst LDR2
    SWP2
    LIST_BIT
    ;fl_cons JSR2       ( slst = cons(x, slst) )
    ,&slst STR2
    JMP2r
&attach
    ;for_sstack JSR2
    ( attach slst to all vars )
    ( var2 addr2 -- )
    POP2
    #0004 ADD2
    DUP2 LDA2 
    DUP2
    ,&slst LDR2
    LIST_BIT
    ;fl_cons JSR2       ( var->cdr = cons(slst, var->cdr) )
    ;fl_addref0 JSR2
    ROT2 STA2
    ;fl_unref JMP2      ( old var->cdr )

    ( resume goals on variable )
    ( var2 -- )
@resume
    DUP2 ,&rsave STR2
    CDR ( get slst )
    ( iterate over slst of var )
&resume1l   ( slst2 -- )
    DUP2 ;fl_nil NEQ2 ,&resume1 JCN
    ( get original slst and unref )
    ,&rsave LDR2 #0004 ADD2 DUP2 LDA2
    ;fl_unref JSR2
    ( store nil in slst of var )
    STA2
    JMP2r
&resume1 ( slst2 -- )
    ( check slst-entry [Goal, Vars/Addrs, ...] of var )
    DUP2 CAR  ( slst2 gva2p )
    ( goal not marked (0)? )
    #0002 ADD2 DUP2 LDA2 ORA ,&lres1 JCN
    POP2
&resume1c ( slst2 -- )
    ( get cdr of slst and continue loop )
    CDR
    ,&resume1l JMP
&rsave $2
&lres1 ( slst2 gva2p -- )
    ( enqueue goal from slst entry )
    LDA2k
    .NGOAL STZ2
    ;enqueue JSR2
    .suspended LDZ2 #0001 SUB2 .suspended STZ2
    #0000 SWP2 STA2k     ( clear goal )
    NIP2 
    #0002 ADD2 LDA2     ( fetch var/addr list )
&resume2l ( slst2 valst2 -- )
    ( iterate over [Var, Addr] pairs )
    DUP2 ;fl_nil NEQ2 ,&lrs2 JCN
    POP2
    ,&resume1c JMP
&lrs2 ( slst2 valst2 -- )
    ( find variable/addr pair in slst )
    DUP2 CAR            ( slst2 valist2 vapair2 )
    #0002 ADD2 LDA2k    ( slst2 valist2 vapair2+2 var2' )
    #fffe AND2          ( unmark var )
    ,&rsave LDR2 EQU2 ,&lres2 JCN
    POP2 CDR
    ,&resume2l JMP
&lres2 ( slst2 valst2 vapair2+2 -- )
    ( set addr of freshly enqueued goal )
    #0002 ADD2 LDA2
    .NGOAL LDZ2 STA2
    POP2
    ,&resume1c JMP

    ( match failed )
@op_fail
    ;EFAIL ;error JMP2

    ( abort with error )
    ( P0=atom2 )
@op_abort
    .PARAM0 LDZ2 
    ;fl_check_atom JSR2
    INC2 ;error JMP2

    ( unify with result )
    ( P0=val1, P1=val2 -> A (true or false) )
@op_unify
    .PARAM0 LDZ2 
    .PARAM1 LDZ2
    ;unify JSR2
    ;op_false JCN2
    ;op_true JMP2
    
    ( unify with check )
    ( P0=val1, P1=val2 )
@op_unify_checked
    .PARAM0 LDZ2 
    .PARAM1 LDZ2
    ;unify JSR2
    #00 EQU ;next JCN2
@eunify
    ;EUNIFY ;error JMP2

    ( unify result with check )
    ( val2a val2b -- )
@fl_unify_result
    ;unify JSR2
    ,eunify JCN
    JMP2r

    ( unify values )
    ( val2a val2b -- f (1 is mismatch) )
@unify
    SWP2 ;fl_deref JSR2
    SWP2 ;fl_deref JSR2
    NEQ2k ,&notsame JCN
    POP2 POP2
    ;false JMP2
&notsame
    ( X != Y )
    OVR2 immediate ,&anonvar JCN
    OVR2 LDA2 VAR_BIT AND2 #0000 EQU2 ,&anonvar JCN
    SWP2 ;fl_assign JSR2
    ;false JMP2
&anonvar
    ( X is not a variable )
    DUP2 immediate ,&fail JCN
    ( Y is not immediate )
    DUP2 LDA2 VAR_BIT AND2 #0000 EQU2 ,&bnonvar JCN
    ;fl_assign JSR2
    ;false JMP2
&bnonvar
    ( Y is not a variable )
    OVR2 immediate ,&fail JCN
    ( X is not immediate )
    OVR2 LDA2 #ff00 AND2 OVR2 LDA2 #ff00 AND2
    EQU2 ,&eqtag JCN
&fail
    POP2 POP2
    ;true JMP2
&eqtag
    ( tags are equal )
    #0002 ADD2 LDA2k SWP2 #0002 ADD2 LDA2 ( X2 Y2CAR Y2CDR )
    ROT2
    #0002 ADD2 LDA2k SWP2 #0002 ADD2 LDA2 ( Y2CAR Y2CDR X2CAR X2CDR )
    ROT2
    ;unify JSR2
    ,&fail JCN
    ;unify JMP2

    ( match values )
    ( G, P0=val1, P1=val2 -> A (true, false or nil) )
@op_match
    .PARAM0 LDZ2
    .PARAM1 LDZ2
    ,match JSR
    ,&lmr1 JCN
    ;op_true JCN2
    ;op_false JMP2
&lmr1
    POP 
@op_nil
    POP2
    ;fl_nil
    ;next JMP2

    ( match values )
    ( G )
    ( val2a val2b -- feq fsusp )
@match
    #02 DEI #04 SUB #03 DEI ;ssave STA2
&match_rec
    SWP2 ;fl_deref JSR2
    SWP2 ;fl_deref JSR2
    NEQ2k ,&noteq JCN
    POP2 POP2
    #0100
    JMP2r
&noteq
    ( not eq )
    OVR2 immediate ,&anovar JCN
    OVR2 LDA2 VAR_BIT AND2 #0000 EQU2 ,&anovar JCN
    POP2
    ( X is var )
&msusp ( var2 )
    ,save_and_suspend JSR
    #0001
&fini
    ;xsave STA2
    ;ssave LDA2 #03 DEO #02 DEO
    ;xsave LDA2
    JMP2r
&anovar
    ( X is not a variable )
    DUP2 immediate #00 EQU ,&bnoimm JCN
&nomatch
    #0000 
    ,&fini JMP
&bnoimm
    DUP2 LDA2 VAR_BIT AND2 ORA ,&msusp JCN
    ( Y is not a variable )
    OVR2 immediate ,&nomatch JCN
    ( X and Y are not immediate and not variables )
    OVR2 LDA2 #ff00 AND2 OVR2 LDA2 #ff00 AND2
    NEQ2 ,&nomatch JCN
    ( tags are equal )
    #0002 ADD2 LDA2k SWP2 #0002 ADD2 LDA2
    ROT2 
    #0002 ADD2 LDA2k SWP2 #0002 ADD2 LDA2
    ROT2
    ;&match_rec JSR2
    POP2
    ;&match_rec JMP2
@ssave $2
@xsave $2

@save_and_suspend ( var2 -- )
    .sstackp LDZ2 STA2k
    NIP2 
    #0002 ADD2
    .GOAL LDZ2 LDA2 SWP2 STA2k
    NIP2
    #0002 ADD2 .sstackp STZ2k
    POP
    goalbuf NEQ2 ,&done JCN
    ;ESSFULL ;error JMP2
&done
    JMP2r

    ( compare for equal identity )
    ( P0, P1 -- F (set if not equal) )
@fl_eq
    .PARAM0 LDZ2 .PARAM1 LDZ2
    NEQ2
    JMP2r

    ( val2 -- int2 )
@op_intval
    ;fl_getint JSR2
    ;next JMP2

    ( int2 -- val2 )
@op_to_int
    ;fl_mkint JSR2
    ;next JMP2

    ( arithmetic )
    ( (SP) + A -> A (int) )
@op_add
    ADD2
    ;next JMP2

    ( (SP) - A -> A (int) )
@op_sub
    SUB2
    ;next JMP2

    ( (SP) * A -> A (int) )
@op_mul
    MUL2
    ;next JMP2

@op_div
    #00 ,divsign STR
    ,divide JSR
    ,divsign LDR #00 EQU ;next JCN2
    #0000 SWP2 SUB2
    ;next JMP2

    ( (SP) % A -> A (int) )
@op_mod
    SWP2 ,abs JSR
    SWP2 ,abs JSR
    DUP2 ORA ,&mod4 JCN
    ;EDIV0 ;error JMP2
&mod4
    DIV2k MUL2 SUB2
    ;next JMP2

    ( (SP) / A -> A (int) )
@divide
    SWP2 ,abs JSR
    SWP2 ,abs JSR
    DUP2 ORA ,&div4 JCN
    ;EDIV0 ;error JMP2
&div4
    DIV2
    JMP2r
@abs ( n2 -- |n2| )
    DUP2 #8000 AND2 POP ,&neg JCN
    JMP2r
&neg
    #0000 SWP2 SUB2
    #01 ,divsign STR
    JMP2r
@divsign $1

    ( integer square root )
    ( A (int) -> A (int) )
@op_sqrt
    ( http://map.grauw.nl/sources/external/z80bits.html#3.2 )
    #ff ,&n STR
    #ffff SWP2
&loop ( R2 A2 )
    OVR2 OVR2 POP NIP #8080 AND2 #8080 EQU2 ,&done JCN
    ADD2k NIP2
    ,&n LDR INC ,&n STR
    SWP2 #01 SUB
    #0001 SUB2
    SWP2
    ,&loop JMP
&done
    POP2
    ,&n LDR #00 SWP
    ;next JMP2
&n $1

    ( binary and )
    ( (SP) & A (int) -> A (int) )
@op_and
    AND2
    ;next JMP2

    ( binary or )
    ( (SP) | A (int) -> A (int) )
@op_or
    ORA2
    ;next JMP2

    ( binary xor )
    ( (SP) ^ A (int) -> A (int) )
@op_xor
    EOR2
    ;next JMP2

    ( binary shift left )
    ( (SP) << A (int) -> A (int) )
@op_shl
    NIP #40 SFT SFT2
    ;next JMP2

    ( binary shift right )
    ( (SP) >> A (int) -> A (int) )
    ( modifies: X0 )
@op_shr
    OVR2 .X0 STZ2
    NIP #0f AND SFT2k ( X2 S1 X2' )
    .X0 LDZ2 #8000 AND2 ORA ,&mask JCN
    ROT POP NIP2
    ;next JMP2
&mask
    ROT #ffff ROT SFT2 #ffff EOR2 ( X2 X2' M2 )
    ORA2 NIP2
    ;next JMP2

    ( match atom or small int )
    ( A=val, <AP>=atomic + goaladdr + jumpaddr )
@op_isval
    DUP2
    ;fetch JSR2
    NEQ2 ;trynext JCN2
    ;skip JSR2
    ;op_skip JMP2

    ( match boxed integer )
    ( A=val, <AP>=int + goaladdr + jumpaddr )
@op_isint
    DUP2
    ;fetch JSR2
    OVR2 immediate ;trynext JCN2
    OVR2 LDA2 INT_BIT AND2 #0000 EQU2 ;trynext JCN2
    OVR2 #0002 LDA2
    NEQ2 ;trynext JCN2
    ;skip JSR2
    ,op_skip JMP

    ( check value for being false and jump if it isn't )
    ( A=val, <AP>=jumpaddr )
@op_isfalse
    DUP2 ;fl_false EQU2 ;op_skip JCN2
    ;op_jump JMP2

@op_skip
    ;skip JSR2
    ;next JMP2

    ( check value for being true and jump if it isn't )
    ( A=val, <AP>(jumpaddr) )
@op_istrue
    DUP2
    ;fl_true EQU2 ;op_skip JCN2
    ;op_jump JMP2

    ( check value for being atomic (small int or atom) )
    ( A=val, <AP>=addr )
@op_isatomic
    DUP2
    immediate ;op_skip JCN2
    ;op_jump JMP2

    ( check value for being list )
    ( val2 -- F (set if not list) )
@fl_islist_t
    DUP2 immediate ,&fail JCN
    LDA2 LIST_BIT AND2 #0000 EQU2
    JMP2r
&fail
    POP2
    #01
    JMP2r
    
    ( check value for being list )
    ( A=val, <AP>=jumpaddr )
@op_islist_t
    DUP2 ,fl_islist_t JSR
    ;op_jump JCN2
    ;op_skip JMP2

    ( check value for being list and deconstruct )
    ( A=val, <AP>=jumpaddr )
@op_islist_n
    DUP2 ,fl_islist_t JSR
    ;op_jump JCN2
    ;skip JSR2
    ;op_decons JMP2

    ( check value for being list and deconstruct )
    ( A=val, <AP>=goaladdr + jumpaddr -> A(car), cdr on stack )
@op_islist
    DUP2 ,fl_islist_t JSR
    ;trynext JCN2
    ;skip JSR2
    ;skip JSR2
    ( fall through ... )

    ( deconstruct cell )
    ( A=val -> A(car), cdr on stack )
@op_decons
    #0004 ADD2 LDA2k 
    SWP2 
    #0002 SUB2 LDA2
    ;next JMP2

    ( check value for being tuple and deconstruct )
    ( A=val, <AP>=arity + goaladdr + jumpaddr -> A (head), arglist on stack )
@op_istuple
    ;fetch JSR2
    OVR2 immediate ,&fail JCN
    OVR2 LDA2 TUPLE_BIT AND2 #0000 EQU2 ,&fail JCN
    OVR2 LDA2 ARITY_MASK AND2 SWP NEQ2 ;trynext JCN2
    ;skip JSR2
    ;skip JSR2
    ,op_decons JMP
&fail
    POP2
    ;trynext JMP2

    ( check for numeric or immediate equality )
    ( P0=val1, P1=val2 -> A )
@op_equal
    .PARAM0 LDZ2
    .PARAM1 LDZ2
    EQU2 ;op_true JCN2
    .PARAM0 LDZ2
    immediate ;op_false JCN2
    .PARAM1 LDZ2
    immediate ;op_false JCN2
    .PARAM0 LDZ2 LDA2 #ff00 AND2
    .PARAM1 LDZ2 LDA2 #ff00 AND2
    NEQ2 ;op_false JCN2
    .PARAM0 LDZ2 INC2 INC2 LDA2
    .PARAM1 LDZ2 INC2 INC2 LDA2
    EQU2 ;op_true JCN2
    ;op_false JMP2

    ( check for numerically less-than )
    ( P0=val1, P1=val2 -> A )
@op_less
    .PARAM0 LDZ2 ;fl_getint JSR2
    .PARAM1 LDZ2 ;fl_getint JSR2
    OVR2 #8000 AND2 OVR2 #8000 AND2 GTH2 ,&sign JCN
    LTH2 ;op_true JCN2
    ;op_false JMP2
&sign
    POP2
    POP2
    ;op_true JMP2
 
    ( check for numerically greater )
    ( P0=val1, P1=val2 -> A )
@op_greater
    .PARAM0 LDZ2 ;fl_getint JSR2
    .PARAM1 LDZ2 ;fl_getint JSR2
    OVR2 #8000 AND2 OVR2 #8000 AND2 LTH2 ,&sign JCN
    GTH2 ;op_true JCN2
    ;op_false JMP2
&sign
    POP2
    POP2
    ;op_true JMP2
 
    ( update goal-ptr and continue at new address )
    ( <AP>=addr + jumpaddr, G )
@trynext
    ;fetch JSR2
    .GOAL LDZ2 STA2
    ;op_jump JMP2

    ( type check for integer )
    ( val -- val )
@fl_check_integer
    DUP2 immediate #00 EQU ,&notimm JCN
    DUP2 lowptr ,&notimm JCN
    JMP2r
&notimm
    DUP2 LDA2 INT_BIT AND2 #0000 EQU2 ,&fail JCN
    JMP2r
&fail
    ;ENOINT ;error JMP2

    ( type check for string )
    ( val -- val )
@fl_check_atom
    DUP2 immediate #00 EQU ,&fail JCN
    DUP2 LOW_BIT AND2 ORA ,&fail JCN
    JMP2r
&fail
    ;ENOSTRING ;error JMP2

    ( type check for cell )
    ( val -- val )
@fl_check_cell
    DUP2 immediate ,&fail JCN
    JMP2r
&fail
    ;ENOCELL ;error JMP2

    ( type check for tuple )
    ( val -- val )
@fl_check_tuple
    DUP2 immediate ,&fail JCN
    DUP2 LDA2 TUPLE_BIT AND2 #0000 EQU2 ,&fail JCN
    JMP2r
&fail
    ;ENOTUPLE ;error JMP2

    ( type check for list )
    ( val -- val )
@fl_check_list
    DUP2 immediate ,&fail JCN
    DUP2 LDA2 LIST_BIT AND2 #0000 EQU2 ,&fail JCN
    JMP2r
&fail
    ;ENOLIST ;error JMP2

    ( type check for var )
    ( val -- val )
@fl_check_var
    DUP2 immediate ,&fail JCN
    DUP2 LDA2 VAR_BIT AND2 #0000 EQU2 ,&fail JCN
    JMP2r
&fail
    ;ENOVAR ;error JMP2

    ( test if tuple )
    ( A=val, <AP>=addr )
@op_istuple_t
    DUP2 immediate ;op_jump JCN2
    DUP2 LDA2 TUPLE_BIT AND2 #0000 EQU2 ;op_jump JCN2
    ;op_skip JMP2

    ( test if int )
    ( P0=val -> A (bool) )
@op_test_number
    .PARAM0 LDZ2 
    DUP2 immediate ,&int1 JCN
    LDA2 INT_BIT AND2 #0000 EQU2 ;op_false JCN2
&int1
    LOW_BIT AND2 #0000 EQU2 ;op_false JCN2
    ;op_true JMP2

    ( test if atom )
    ( P0=val -> A (bool) )
@op_test_atom
    .PARAM0 LDZ2
    DUP2 immediate ,&atm1 JCN
    POP2
    ;op_false JMP2
&atm1
    LOW_BIT AND2 #0000 EQU2 ;op_true JCN2
    ;op_false JMP2

    ( test if var )
    ( P0=val -> A (bool) )
@op_test_var
    .PARAM0 LDZ2
    DUP2 immediate ,&var1 JCN
    LDA2 VAR_BIT AND2 #0000 EQU2 ;op_false JCN2
    ;op_true JMP2
&var1
    POP2
    ;op_false JMP2

    ( test if tuple )
    ( P0=val -> A (bool) )
@op_test_tuple
    .PARAM0 LDZ2
    DUP2 immediate ,&tup1 JCN
    LDA2 TUPLE_BIT AND2 #0000 EQU2 ;op_false JCN2
    ;op_true JMP2
&tup1
    POP2
    ;op_false JMP2

    ( test if list )
    ( P0=val -> A (bool) )
@op_test_list
    .PARAM0 LDZ2
    DUP2 immediate ,&lst1 JCN
    LDA2 LIST_BIT AND2 #0000 EQU2 ;op_false JCN2
    ;op_true JMP2
&lst1
    POP2
    ;op_false JMP2

    ( absolute value )
    ( A (int) -> A (int) )
@op_abs
    DUP2 #8000 AND2 #0000 EQU2 ;next JCN2
    #0000 SWP2 SUB2
    ;next JMP2

    ( sign of integer )
    ( A=val (int) -> A (int) )
@op_sign
    DUP2 ORA ,&nonzero JCN
    ;next JMP2
&nonzero
    POP #80 AND ,&negative JCN
    #0001
    ;next JMP2
&negative
    #ffff
    ;next JMP2

    ( ordered compare )
    ( P0=val1, P1=val2 -> A (bool or nil) )
@op_below
    .PARAM0 LDZ2
    .PARAM1 LDZ2
    ;ordering JSR2
    ,ofail JCN
    #ff EQU ;op_true JCN2
    ;op_false JMP2
@ofail
    POP
    POP2 ;fl_nil 
    ;next JMP2

    ( P0=val1, P1=val2 -> A (bool) )
@op_above
    .PARAM0 LDZ2
    .PARAM1 LDZ2
    ;ordering JSR2
    ,ofail JCN
    #01 EQU ;op_true JCN2
    ;op_false JMP2

    ( access idle var )
    ( <AP>=cache -> A (bool) )
@op_idle
    ;fetch JSR2
    LDAk INC
    .epoch LDZ NEQ ,&wait JCN
    #0000 SWP2 STA2
    ;op_true JMP2
&wait
    .epoch LDZ #00 SWP2 STA2
    .idle LDZ2
    ;sspush JSR2
    POP
    ;op_false JMP2

    ( maximum value )
    ( max((SP), A) -> A (int) )
@op_max
    OVR2 OVR2 EOR2 POP #80 AND ,&maxsign JCN
    GTH2k JMP SWP2 POP2
    ;next JMP2
&maxsign
    OVR #07 SFT JMP SWP2 POP2
    ;next JMP2

    ( minimum value )
    ( min((SP), A) -> A (int) )
@op_min
    OVR2 OVR2 EOR2 POP #80 AND ,&minsign JCN
    LTH2k JMP SWP2 POP2
    ;next JMP2
&minsign
    OVR2 POP #07 SFT JMP SWP2 POP2
    ;next JMP2

    ( allocate list cell )
    ( A=cdr, (sp + 2)=car -> A )
@op_mklist
    SWP2 LIST_BIT ;fl_cons JSR2
    ;next JMP2

    ( allocate tuple )
    ( A=args, <AP>=atom + arity -> A )
@op_mktuple
    ;fetch JSR2
    ;fetch JSR2
    #80 SFT2
    TUPLE_BIT ORA2
    ;fl_cons JSR2
    ;next JMP2

    ( mutate list into general tuple )
    ( lst2 arity1 -- tup2 )
@fl_mktuple0
    DUP #f8 AND #00 EQU ,&mkt JCN
    ;EARITY ;error JMP2
&mkt
    #40 ORA     ( TUPLE_BIT )
    ROT ROT STAk ROT POP
    JMP2r

@op_mktuple0
    ;fetch JSR2 NIP
    ,fl_mktuple0 JSR
    ;next JMP2

    ( threaded interpreter )
@next
    .AP LDZ2 
    ( #3a debug )
    LDA2k
    SWP2 INC2 INC2 .AP STZ2
    JMP2

    ( -- addr2 )
@fetch
    .AP LDZ2 LDA2k
    SWP2 INC2 INC2 .AP STZ2
    JMP2r

@skip
    .AP LDZ2 INC2 INC2 .AP STZ2
    JMP2r

    ( rst: points to threaded code )
@fl_exec
    .AP STH STZ2r
    ,next JMP

    ( unconditional addref, expects block )
    ( A=val )
@op_addref
    DUP2 LDA2k INC SWP2 STA2
    ;next JMP2

    ( A=val )
@op_addrefx
    ;fl_addref JSR2
    ;next JMP2

    ( P0=val, A=var )
@op_assign
    .PARAM0 LDZ2
    OVR2
    ;fl_assign JSR2
    ;next JMP2

    ( test if atomic )
    ( P0=val -> A (bool) )
@op_atomic
    .PARAM0 LDZ2
    DUP2 immediate ;op_true JCN2
    LDA2 INT_BIT AND2 ORA ;op_true JCN2
    ( fall through ... )

    ( -> A )
@op_false
    POP2
    ;fl_false
    ;next JMP2

    ( -> A )
@op_true
    POP2
    ;fl_true
    ;next JMP2

    ( <AP>=addr -> A )
@op_peek
    POP2
    ;fetch JSR2
    LDA2
    ;next JMP2

    ( <AP>=addr -> A )
@op_peek_deref
    POP2
    ;fetch JSR2
    LDA2
    ;op_deref JMP2

    ( A, <AP>=addr )
@op_poke
    DUP2
    ;fetch JSR2
    STA2
    ;next JMP2

    ( A, <AP>=addr )
@op_store
    ;fl_addref JSR2
    ,op_poke JMP
    
    ( <AP>=addr, mkvar -> A )
@op_storevar
    POP2
    ;fl_mkvar JSR2
    ,op_store JMP
    
    ( A=val )
@op_check_atom
    ;fl_check_atom JSR2
    ;next JMP2
@op_check_cell
    ;fl_check_cell JSR2
    ;next JMP2
@op_check_integer
    ;fl_check_integer JSR2
    ;next JMP2
@op_check_list
    ;fl_check_list JSR2
    ;next JMP2
@op_check_tuple
    ;fl_check_tuple JSR2
    ;next JMP2
@op_check_var
    ;fl_check_var JSR2
    ;next JMP2

@op_intval_checked
    ;fl_check_integer JSR2
    ;op_intval JMP2

@op_tryelse
    reset_stack
    #0000 ( dummy A )
    ;next JMP2

    ( <AP> -> A )
@op_lit
    POP2
    ;fetch JSR2
    ;next JMP2

    ( <AP> -> A )
@op_litlabel
    POP2
    ;fetch JSR2
    #8001 ORA2
    ;next JMP2

    ( A -> A )
@op_deref
    ;fl_deref JSR2
    ;next JMP2

    ( A -> A, <AP>=jumpaddr )
@op_force
    ;fl_force JSR2
    ;op_jump JCN2
    ;op_skip JMP2

    ( <AP>=jumpaddr )
@op_forcearg
    .GOAL LDZ2 #0002 ADD2 LDA2k
    ;fl_force JSR2
    ,&jump JCN
    OVR2 LDA2 OVR2 EQU2 ,&skip JCN
    ;fl_addref JSR2
    SWP2 DUP2 LDA2 ;fl_unref JSR2
    STA2
    ;op_skip JMP2
&jump
    POP2 POP2
    ;op_jump JMP2
&skip
    POP2 POP2
    ;op_skip JMP2
    
    ( <AP>=addr )
@op_jump
    ;fetch JSR2
    .AP STZ2
    ;next JMP2

    ( <AP>=addr )
@op_fcall
    ;fetch JSR2
    ;next STH2
    JMP2

    ( -> A )
@op_getarg1
    #0002
@geta
    .GOAL LDZ2 ADD2 LDA2 
    NIP2
    ;next JMP2    
@op_getarg2
    #0004
    ,geta JMP
@op_getarg3
    #0006
    ,geta JMP
@op_getarg4
    #0008
    ,geta JMP

    ( A )
@op_putarg1
    #0002
@puta
    .NGOAL LDZ2 ADD2 STA2k POP2
    ;next JMP2
@op_putarg2
    #0004
    ,puta JMP
@op_putarg3
    #0006
    ,puta JMP
@op_putarg4
    #0008
    ,puta JMP

    ( A )
@op_storearg1
    #0002
@stoa
    .NGOAL LDZ2 ADD2 STA2k POP2
    ;fl_addref JSR2
    ;next JMP2
@op_storearg2
    #0004
    ,stoa JMP
@op_storearg3
    #0006
    ,stoa JMP
@op_storearg4
    #0008
    ,stoa JMP

    ( A )
@op_putparam1
    DUP2 .PARAM0 STZ2
    ;next JMP2
@op_putparam2
    DUP2 .PARAM1 STZ2
    ;next JMP2
@op_putparam3
    DUP2 .PARAM2 STZ2
    ;next JMP2
@op_putparam4
    DUP2 .PARAM3 STZ2
    ;next JMP2

    ( <AP>=addr )
@op_not_suspending
    .sstackp LDZ2 sstack EQU2 ;op_jump JCN2
    ;op_skip JMP2
@op_suspending
    .sstackp LDZ2 sstack NEQ2 ;op_jump JCN2
    ;op_skip JMP2

    ( P0, P1, <AP>=jumpaddr )
@op_guard
    ;fl_eq JSR2
    ;op_jump JCN2
    ;op_skip JMP2

    ( <AP>=addr )
@op_opengoal
    ;fl_opengoal JSR2
    ;fetch JSR2
    .NGOAL LDZ2 STA2
    ;next JMP2

    ( -> A )
@op_pop
    POP2
    ;next JMP2

    ( A )
@op_push
    DUP2
    ;next JMP2

@op_reset
    sstack .sstackp STZ2
    ;next JMP2

    ( <AP>=addr )
@op_drop
    ;fetch JSR2
    LDA2
    ;fl_unref JSR2
    ;next JMP2

    ( <AP>=addr )
@op_drop2
    ;fetch JSR2
    LDZ2
    ;fl_unref JSR2
    LDZ2
    ;fl_unref JSR2
    ;next JMP2

@op_droparg1
    #0002
@dropa
    .GOAL LDZ2 ADD2 LDA2 ;fl_unref JSR2
    ;next JMP2
@op_droparg2
    #0004
    ,dropa JMP
@op_droparg3
    #0006
    ,dropa JMP
@op_droparg4
    #0008
    ,dropa JMP

    ( auxilliary routines )

    ( ordering of types )
    ( val2 -- ordercode )
@typeorder
    DUP2 immediate ,&imm JCN
    LDA2
    DUP2 TUPLE_BIT AND2 #0000 EQU2 ,&nontup JCN
    POP2
    #04
    JMP2r
&nontup
    DUP2 VAR_BIT AND2 #0000 EQU2 ,&nonvar JCN
    POP2
    #00
    JMP2r
&nonvar
    INT_BIT AND2 #0000 EQU2 ,&nonint JCN
    #01
    JMP2r
&nonint
    #03
    JMP2r
&imm
    lowptr ,&str JCN
    #01 
    JMP2r
&str
    #02
    JMP2r

    ( ordering of values )
    ( val2a val2b -- sign fsusp )
@ordering
    #02 DEI #04 SUB #03 DEI ;ssave STA2
&order_rec ( val2a val2b -- sign fsusp )
    OVR2 OVR2 ,typeorder JSR
    DUP #00 EQU ,&bvar JCN      ( val2a val2b val2a tb )
    ROT ROT ,typeorder JSR      ( val2a val2b tb ta )
    DUP #00 EQU ,&avar JCN
    SWP
    DUP2 SUB                    ( val2a val2b td )
    DUP #00 EQU ,&diff JCN
    #00
&result ( diff2 -- sign 0 )
    ,&sign JSR
&result0 ( sign )
    #00 
&fini ( sign fsusp )
    ;xsave STA2
    ;ssave LDA2 #03 DEO #02 DEO
    ;xsave LDA2
    JMP2r
&sign ( n2 -- s )
    DUP2 ORA ,&nonzero JCN
    POP2
    #00 JMP2r
&nonzero
    #8000 AND2 #0000 EQU2 ,&pos JCN
    #ff JMP2r
&pos
    #01 JMP2r
&bvar ( val2a val2b val2a tb -- )
    NIP NIP
&avar ( var2a ta -- )
    POP
    ;save_and_suspend JSR2
    #0001
    ,&fini JMP
&diff ( val2a val2b ta1 tb1 td1 )
    POP2
    #10 SFT
    #00 SWP
    ;&cmptable ADD2
    LDA2 JMP2
&cmptable
    :&cmpptr
    :&cmpint
    :&cmpstr
    :&cmplst
    :&cmptup
&cmpint
    SWP2 ;fl_getint JSR2
    SWP2 ;fl_getint JSR2
    SUB2
    ,&sign JSR
    #00
    JMP2r
&cmpptr
    SUB2
    ,&sign JSR
    #00 
    JMP2r
&cmplst
    OVR2 #0002 ADD2 LDA2
    OVR2 #0002 ADD2 LDA2
    ;&order_rec JSR2
    POP DUP ;&result0 JCN2
    POP
    SWP2 #0004 ADD2 LDA2
    SWP2 #0004 ADD2 LDA2
    ;&order_rec JMP2
&cmptup
    OVR2 LDA2 #ff00 AND2
    OVR2 LDA2 #ff00 AND2
    SUB2 DUP2
    ORA ;&result JCN2
    POP2
    ,&cmplst JMP
&cmpstr
    OVR2 LDA #00 SWP
    OVR2 LDA #00 SWP
    SUB2 DUP2
    ORA ;&result JCN2
    POP2
    DUP2 LDA ,&cont JCN
    POP2 POP2
    #0000
    JMP2r
&cont
    SWP2 INC2
    SWP2 INC2
    ,&cmpstr JMP

    ( E<I> -> Arg<J> -> A, <AP>=[I,J] (already multiplied by 2) )
@op_move_e_a
    POP2
    ;fetch JSR2
    SWP .ENV ADD LDZ2
    ROT #00 SWP
    .NGOAL LDZ2 ADD2 STA2k
    POP2
    ;op_addrefx JMP2

    ( mkvar -> E<I> -> Arg<J> -> A, <AP>=[I,J] (already multiplied by 2) )
@op_move_var_e_a
    POP2
    ;fetch JSR2
    SWP
    ;fl_mkvar JSR2
    ;fl_addref0 JSR2        ( J I var2 )
    ROT .ENV ADD STZ2k POP  ( J var2 )
    ROT #00 SWP
    .NGOAL LDZ2 ADD2 STA2k POP2
    ;op_addrefx JMP2

    ( Arg<I> -> E<J> -> A, <AP>=[I,J] (already multiplied by 2) )
@op_move_a_e
    POP2
    ;fetch JSR2
    #00 ROT .GOAL LDZ2 ADD2 LDA2
    ROT .ENV ADD STZ2k
    POP
    ;op_addrefx JMP2

    ( E<I> -> P<J> -> A, <AP>=[I,J] (already multiplied by 2) )
@op_move_e_p
    POP2
    ;fetch JSR2
    SWP .ENV ADD LDZ2
    ;fl_deref JSR2
    ROT .PARAM0 ADD STZ2k
    POP
    ;next JMP2

    ( mkvar -> E<I> -> P<J> -> A, <AP>=[I,J] (already multiplied by 2) )
@op_move_var_e_p
    POP2
    ;fetch JSR2
    SWP
    ;fl_mkvar JSR2
    ;fl_addref0 JSR2        ( J I var2 )
    ROT .ENV ADD STZ2k POP  ( J var2 )
    ROT .PARAM0 ADD STZ2k
    POP
    ;next JMP2

    ( A, <AP> = [Count, Val, Label, ...] )
@op_lookup
    .X0 STZ2
    ;fetch JSR2
    NIP
&lok1 ( count -- )
    ;fetch JSR2
    .X0 LDZ2
    NEQ2 ,&next JCN
&done
    POP
    .X0 LDZ2
    ;op_jump JMP2
&next
    ;skip JSR2
    #01 SUB
    DUP ,&lok1 JCN
    ,&done JMP

    ( A ENV-index -- A )
@fl_xassign
    NIP .ENV ADD LDZ2
    ;fl_assign JMP2

    ( iterate over sstack )
    ( (SP)=addr, invoke addr with entries: var addr )
    ( (assumes sstack is not empty) )
@for_sstack
    sstack
&loop
    DUP2 LDA2
    OVR2 INC2 INC2 LDA2
    ;&next STH2 OVR2r JMP2r
&next
    #0004 ADD2
    DUP2 .sstackp LDZ2 NEQ2 ,&loop JCN
    POP2
    POP2r
    JMP2r

    ( write hex number )
    ( num2 )
@writehex
    SWP
    ,&wbyte JSR
&wbyte
    DUP
    #04 SFT
    ,&wdigit JSR
    #0f AND
&wdigit
    #0a LTHk ,&digit JCN
    POP
    #57 ADD
    #18 DEO
    JMP2r
&digit
    POP
    #30 ADD
    #18 DEO
    JMP2r

    ( copy memory )
    ( from2 to2 n2 -- )
@fl_copy
    SWP2 STH2
&loop
    DUP2 ORA ,&next JCN
    POP2r POP2 POP2
&next
    SWP2 LDAk STH2kr STA
    INC2 SWP2 #0001 SUB2 
    ,&loop JMP

    ( vector hook )
@vector ( vec )
    ( #7c debug )
    DUP #02 ADD STH
    LDZ2k INC2 ROT STZ2k   ( inc counter )
    POP ;fl_mkint JSR2
    ;fl_mkvar JSR2 SWP2 LIST_BIT
    ;fl_cons JSR2
    DUP2
    STHkr LDZ2 ;chase_tail JSR2
    STHr STZ2
    ;op_yield JMP2
    
@vector00 .vectors ;vector JMP2
@vector10 .vectors #04 ADD ;vector JMP2
@vector20 .vectors #08 ADD ;vector JMP2
@vector30 .vectors #0c ADD ;vector JMP2
@vector40 .vectors #10 ADD ;vector JMP2
@vector50 .vectors #14 ADD ;vector JMP2
@vector60 .vectors #18 ADD ;vector JMP2
@vector70 .vectors #1c ADD ;vector JMP2
@vector80 .vectors #20 ADD ;vector JMP2
@vector90 .vectors #24 ADD ;vector JMP2
@vectora0 .vectors #28 ADD ;vector JMP2
@vectorb0 .vectors #2c ADD ;vector JMP2
@vectorc0 .vectors #30 ADD ;vector JMP2
@vectord0 .vectors #34 ADD ;vector JMP2
@vectore0 .vectors #38 ADD ;vector JMP2
@vectorf0 .vectors #3c ADD ;vector JMP2

@vector_table
:vector00
:vector10
:vector20
:vector30
:vector40
:vector50
:vector60
:vector70
:vector80
:vector90
:vectora0
:vectorb0
:vectorc0
:vectord0
:vectore0
:vectorf0

    ( chase stream tail and assign val )
    ( X2 S2 -- )
@chase_tail
    STH2k
&loop
    ;fl_deref JSR2
    DUP2 LDA2 LIST_BIT AND2 #0000 EQU2 ,&end JCN
    ( loop until non-list (expected: var) )
    #0004 ADD2 LDA2
    ,&loop JMP
&end
    OVR2 ;fl_addref JSR2 POP2 ( addref tail )
    STH2r ;fl_unref JSR2 ( unref old S )
    ;fl_assign JMP2  ( assign X to tail )

    ( foreign calls )

    ( convert integer to list )
    ( P0=num, P1=list, P2=tail )
    ( modifies: X0 )
@fl_number_to_list_3
    #00 ,&isneg STR
    #00 ,&keep STR
    .PARAM0 LDZ2 ;fl_getint JSR2
    ;&buf .X0 STZ2
    DUP2 ORA ,&non0 JCN
    #30 ,&buf STR
    ;&buf INC2 .X0 STZ2
    ,&consbuf JMP
&non0
    OVR #80 AND #00 EQU ,&nonneg JCN
    #01 ,&isneg STR
    #0000 SWP2 SUB2
&nonneg
    ,&num2dec JSR
    POP
&consbuf
    .PARAM2 LDZ2
    .X0 LDZ2
&loop ( tl2 ptr2 )
    #0001 SUB2
    LDAk #40 SWP #10 SFT2 #0001 ORA2    ( tl2 ptr2 n2 )
    ROT2 SWP2
    LIST_BIT
    ;fl_cons JSR2
    SWP2
    DUP2 ;&buf NEQ2 ,&loop JCN
    POP2
    ,&isneg LDR #00 EQU ,&done JCN
    #805b LIST_BIT ;fl_cons JSR2
&done
    .PARAM1 LDZ2
    ;fl_unify_result JMP2
&isneg $1
&keep $1
&buf $8
    ( taken from helpers.tal: )
&num2dec ( short* -- )
    #03e8 DIV2k
    DUP ,&second JSR
    MUL2 SUB2
    #0064 DIV2k
    DUP ,&third JSR
    MUL2 SUB2
&second
    #0a DIVk DUP
    ,&third JSR
    MUL SUB
&third
    DUP ,&keep LDR ADD #00 EQU ,&next JCN
    #01 ,&keep STR
    #30 ADD
    .X0 LDZ2 STAk INC2 .X0 STZ2
&next
    POP
    JMP2r

    ( list length )
    ( list2 -- len1 )
@llength
    #00 .X0 STZ
&llen1
    ;fl_deref JSR2
    DUP2 immediate ,&done JCN
    DUP2 LDA2 LIST_BIT AND2 #0000 EQU2 ,&done JCN
    .X0 LDZ INC .X0 STZ
    #0004 ADD2 LDA2
    ,&llen1 JMP
&done
    POP2
    .X0 LDZ
    JMP2r

@fl_assign_3
    .PARAM0 LDZ2
    .PARAM1 LDZ2
    ;fl_assign JSR2
    .PARAM2 LDZ2
    ( fall through ... )

    ( var -- )
@confirm
    ;fl_nil SWP2
    ;fl_assign JMP2

@fl_list_to_tuple_2
    .PARAM0 LDZ2
    DUP2
    ,llength JSR
    #01 SUB
    ;fl_mktuple0 JSR2
    .PARAM1 LDZ2
    ;fl_unify_result JMP2

@fl_halt_0
    ;exit JMP2

@fl_eload
    ;ELOAD ;error JMP2

@fl_uncons_3
    .PARAM0 LDZ2
    ;fl_check_cell JSR2
    #0002 ADD2 LDA2k
    .PARAM1 LDZ2
    ;fl_assign JSR2
    #0002 ADD2 LDA2
    .PARAM2 LDZ2
    ;fl_assign JMP2

@fl_string_to_list_3
    .PARAM0 LDZ2
    ;fl_check_atom JSR2
    LDAk #00 SWP OVR2 ADD2
    .PARAM2 LDZ2 .X0 STZ2
&loop
    EQU2k ,&done JCN
    LDAk 
    #40 SWP
    #10 SFT2
    #01 ORA
    .X0 LDZ2 SWP2
    LIST_BIT 
    ;fl_cons JSR2
    .X0 STZ2
    #0001 SUB2
    ,&loop JMP
&done
    POP2 POP2
    .X0 LDZ2
    .PARAM1 LDZ2
    ;fl_unify_result JMP2

    ( locate and check index into tuple )
    ( tup2 index1 -- X2 )
@index
    #01 SUB
    DUP #00 EQU ,&ind2 JCN
    DUP #f8 ( BITS_MASK ) AND
    #00 EQU ,&ind1 JCN
    ;EBADINDEX ;error JMP2
&ind1
    ROT ROT
    #0004 ADD2 LDA2
    ;fl_deref JSR2
    ROT #01 SUB
    DUP ,&ind1 JCN
&ind2
    POP
    #0002 ADD2 LDA2
    JMP2r

@fl_get_arg_3
    .PARAM0 LDZ2
    ;fl_getint JSR2
    NIP
    .PARAM1 LDZ2
    ;fl_check_tuple JSR2
    ROT ,index JSR
    .PARAM2 LDZ2
    ;fl_unify_result JMP2

@fl_put_arg_3
    .PARAM0 LDZ2
    ;fl_getint JSR2
    NIP
    .PARAM1 LDZ2
    ;fl_check_tuple JSR2
    ROT ,index JSR
    .PARAM2 LDZ2 
    SWP2 
    ;fl_assign JMP2

    ( P0=[S|_], P1=X - chase tail in S and assign X )
@fl_merge_2
    .PARAM1 LDZ2
    .PARAM0 LDZ2
    #0002 ADD2 LDA2                ( get S )
    ;chase_tail JSR2
    .PARAM1 LDZ2 
    .PARAM0 LDZ2 #0002 ADD2 STA2  ( destructively change S to X )
    JMP2r

    ( P0=[S|Count], P1=Amount )
@fl_merge_refs_2
    .PARAM0 LDZ2
    #0005 ADD2 LDAk #01 SUB
    .PARAM1 INC LDZ ADD 
    ROT ROT
    STA
    JMP2r
    
    ( P0=addr, P1=byte )
@fl_poke_2
    .PARAM1 LDZ2
    ;fl_check_integer JSR2
    ;fl_getint JSR2
    NIP
    ,getaddr JSR
    STA
    JMP2r

@getaddr ( -- addr2 )
    .PARAM0 LDZ2
    ;fl_check_integer JSR2
    ;fl_getint JMP2

    ( P0=addr, P1=result )
@fl_peek_2
    ,getaddr JSR
    LDA
    #40 SWP 
    #10 SFT2
    #01 ORA
    .PARAM1 LDZ2
    ;fl_assign JMP2

    ( fetch system vector stream )
    ( P0=vec, P1=S )
@fl_get_vector_2
    .PARAM0 LDZ2
    ;fl_check_integer JSR2
    ;fl_getint JSR2
    NIP
    #20 SFT .vectors ADD INC INC LDZ2
    DUP2 ORA #00 EQU ,&done JCN
    POP2 ;fl_nil
&done    
    .PARAM1 LDZ2 ;fl_assign JSR2
    JMP2r

    ( assign to system vector )
    ( P0=vec, P1=S )
@fl_set_vector_2
    .PARAM0 LDZ2
    ;fl_check_integer JSR2
    ;fl_getint JSR2
    NIP STHk
    ( reset counter, create stream )
    #20 SFT .vectors ADD #0000 ROT STZ2k
    #02 ADD ;fl_mkvar JSR2 ;fl_addref JSR2 
    ROT STZ2k
    POP
    .PARAM1 LDZ2 ;fl_assign JSR2
    POP2
    ( setup device )
    STHrk #10 SFT #00 SWP ;vector_table ADD2 LDA2
    STHr #40 SFT DEO2
    .hooked LDZk INC SWP STZ
    JMP2r

    ( reset vector )
    ( P0=vec )
@fl_reset_vector_1
    .PARAM0 LDZ2
    ;fl_check_integer JSR2
    ;fl_getint JSR2
    NIP
    DUP #40 SFT #0000 ROT DEO
    #20 SFT .vectors ADD #0000 ROT STZ2k
    #02 ADD LDZ2k ;fl_deref JSR2
    POP2
    .hooked LDZk #01 SUB SWP STZ
    JMP2r

( data area )

@ENOCELLS "HEAP 20 "FULL 00
@ENOGOALS "GOAL 20 "BUFFER 20 "FULL 00
@EQUEUEFULL "PROCESS 20 "POOL 20 "OVERFLOW 00
@EBOUND "ASSIGNMENT 20 "TO 20 "NON-VARIABLE 00
@EFAIL "FAILED 00
@EUNIFY "CAN 20 "NOT 20 "UNIFY 00
@EDIV0 "DIVISION 20 "BY 20 "ZERO 00
@EDEADLOCK "DEADLOCK 00
@ELEAK "MEMORY 20 "LEAK 00
@ENOINT "EXPECTED 20 "INTEGER 00
@ENOSTRING "EXPECTED 20 "STRING 00
@ENOVAR "EXPECTED 20 "VARIABLE 00
@ENOTUPLE "EXPECTED 20 "TUPLE 00
@ENOLIST "EXPECTED 20 "LIST 00
@ENOCELL "EXPECTED 20 "CELL 00
@EBADINDEX "TUPLE 20 "INDEX 20 "OUT 20 "OF 20 "RANGE 00
@ESSFULL "SUSPENSION-STACK 20 "OVERFLOW 00
@EARITY "TUPLE 20 "LENGTH 20 "MAXIMUM 20 "EXCEEDED 00
@ELOAD "FILE 20 "OPERATION 20 "FAILED 00

* 00 @fl_nil 02 "[] 00
* 00 @fl_true 04 "true 00
* 00 @fl_false 05 "false 00

@init
