% FLENG compiler - Uxn backend

:- dynamic(fragment/3).

target_architecture(uxn).

init_asm(Infile, Out) :-
    fleng_version(V),
    register_atom('[]', 0),
    register_atom('true', 1),
    register_atom('false', 2),
    resolve_addr(init, A),
    hex(A, Ax),
    clear_a,
    g_assign(xlabel, 0),
    target_architecture(Arch),
    format(Out, '( GENERATED BY fl2~a [~a] FROM ~a )\n~~uxn.inc\n|~a ;fleng_init JMP2\n', 
        [Arch, V, Infile, Ax]).

clear_a :- g_assign(a_current, none).
set_a(R) :- g_assign(a_current, R).
assigned_a(R) :- g_read(a_current, R).

asm([], _).
asm([Instr|Asm], Out) :-
    asm(Instr, Out),
    !,
    asm(Asm, Out).

asm(above, Out) :- t_op(above, Out).
asm(addr_of(Label), Out) :- 
    clear_a, 
    t_op(lit, Out),
    emit('  :~a\n', [Label], Out).
asm(addref, Out) :- t_op(addref, Out).
asm(addrefx, Out) :- t_op(addrefx, Out).
asm(asm(Code), Out) :- 
    emit('~a\n', [Code], Out).
asm(asm_enter, Out) :- 
    g_read(xlabel, N),
    N2 is N + 1,
    g_assign(xlabel, N2),
    emit('  :asml~d @asml~d\n', [N, N], Out).
asm(asm_exit, Out) :- 
    clear_a,
    resolve_addr(fl_exec, Exec),
    hex(Exec, Exec2),
    emit('  #~a JSR2 ( fl_exec )\n', [Exec2], Out).
asm(assign, Out) :- 
    clear_a, 
    t_op(assign, Out).
asm(atom(I, S), Out) :-
    resolve_atom_index(I, A),
    clear_a,
    t_op(lit, Out),
    emit('  ~a ( \'~a\' )\n', [A, S], Out).
asm(atomic, Out) :- 
    clear_a,
    t_op(atomic, Out).
asm(below, Out) :- t_op(below, Out).
asm(bind(I), Out) :- 
    resolve_addr('ENV', A),
    P is A + I * 2,
    hex(P, PX),
    set_a(e(I)),
    t_op(store, Out),
    emit('  ~a ( ENV~d )\n', [PX, I], Out).
asm(bindvar(I), Out) :- 
    resolve_addr('ENV', A),
    P is A + I * 2,
    hex(P, PX),
    set_a(e(I)),
    t_op(storevar, Out),
    emit('  ~a ( ENV~d )\n', [PX, I], Out).
asm(check_atom, Out) :- t_op(check_atom, Out).
asm(check_cell, Out) :- t_op(check_cell, Out).
asm(check_integer, Out) :- t_op(check_integer, Out).
asm(check_list, Out) :- t_op(check_list, Out).
asm(check_tuple, Out) :- t_op(check_tuple, Out).
asm(check_var, Out) :- t_op(check_var, Out).
asm(clear_n, _).
asm(code_label(N/A, I), Out) :-
    t_op(litlabel, Out),
    clear_a,
    emit('  :P~d ( label: ~a/~d )\n', [I, N, A], Out).
asm(compute(+), Out) :- clear_a, t_op(add, Out).
asm(compute(-), Out) :- clear_a, t_op(sub, Out).
asm(compute(*), Out) :- clear_a, t_op(mul, Out).
asm(compute(/), Out) :- clear_a, t_op(div, Out).
asm(compute(mod), Out) :- clear_a, t_op(mod, Out).
asm(compute(and), Out) :- clear_a, t_op(and, Out).
asm(compute(or), Out) :- clear_a, t_op(or, Out).
asm(compute(xor), Out) :- clear_a, t_op(xor, Out).
asm(compute(shl), Out) :- clear_a, t_op(shl, Out).
asm(compute(shr), Out) :- clear_a, t_op(shr, Out).
asm(compute(=), Out) :- clear_a, t_op(equal, Out).
asm(compute(<), Out) :- clear_a, t_op(less, Out).
asm(compute(>), Out) :- clear_a, t_op(greater, Out).
asm(compute(sign), Out) :- clear_a, t_op(sign, Out).
asm(compute(abs), Out) :- clear_a, t_op(abs, Out).
asm(compute(sqrt), Out) :- clear_a, t_op(sqrt, Out).
asm(compute(max), Out) :- clear_a, t_op(max, Out).
asm(compute(min), Out) :- clear_a, t_op(min, Out).
asm(data(L), Out) :- emit('@L~d 0000\n', [L], Out).
asm(debug, Out) :- clear_a, t_op(dbg, Out).
asm(defatom(I, A), Out) :-
    I >= 3,
    string_length(A, Len),
    emit('  * 00 @S~d ', [I], Out),
    atom_codes(A, C),
    emit_list([Len|C], Out),
    emit(' 00\n', Out).
asm(defatom(_, _), _).
asm(define(A, I, _, N), Out) :-
    % assumes: G, SP are set
    emit('  * ( ~a/~d )\n@P~d\n', [N, A, I], Out).
asm(deref, Out) :- t_op(deref, Out).
asm(drop(I), Out) :- 
    resolve_addr('ENV', A),
    t_op(drop, Out),
    N is A + I * 2,
    hex(N, NX),
    emit('  ~a ( ENV~d )\n', [NX, I], Out).
asm(drop2(I, J), Out) :- 
    resolve_addr('ENV', A),
    t_op(drop2, Out),
    N is ((A + I * 2) << 8) \/ (A + J *  2),
    hex(N, NX),
    emit('  ~a ( ENV~d, ENV~d )\n', [NX, I, J], Out).
asm(droparg(I), Out) :-
    I2 is I + 1,
    t_op(droparg, I2, Out).
asm(dropgoal, Out) :- t_op(dropgoal, Out).
asm(end, Out) :- 
    resolve_addr(fl_initsys, A),
    hex(A, AX),
    emit('@fleng_init  #~a JSR2 ( fl_initsys )\n', [AX], Out),
    emit_entry_point(Out),
    emit('( END OF FILE )\n', Out).
asm(error, Out) :- t_op(abort, Out).
asm(fail(_), Out) :- t_op(fail, Out).
asm(force(I), Out) :- 
    t_op(force, Out),
    emit('  :L~d\n', [I], Out).
asm(force_arg(L), Out) :- 
    t_op(forcearg, Out),
    emit('  :L~d\n', [L], Out).
asm(foreign_call(N, A), Out) :-
    t_op(fcall, Out),
    format_to_atom(Call, '~a_~d', [N, A]),
    resolve_addr(Call, Addr),
    hex(Addr, X),
    clear_a,
    emit('  ~a ( ~a_~d )\n', [X, N, A], Out).
asm(getarg(I), Out) :- 
    \+assigned_a(a(I)),
    I2 is I + 1,
    set_a(a(I)),
    format_to_atom(Op, 'getarg~d', [I2]),
    t_op(Op, Out).
asm(getarg(_), _).
asm(getenv(I), Out) :- 
    \+assigned_a(e(I)),
    resolve_addr('ENV', A),
    N is A + I * 2,
    hex(N, NX),
    set_a(e(I)),
    t_op(peek, Out),
    emit('  ~a ( ENV~d )\n', [NX, I], Out).
asm(getenv(_), _).
asm(getenv_deref(I), Out) :- 
    \+assigned_a(e(I)),
    resolve_addr('ENV', A),
    N is A + I * 2,
    hex(N, NX),
    set_a(e(I)),
    t_op(peek_deref, Out),
    emit('  ~a ( ENV~d )\n', [NX, I], Out).
asm(getenv_deref(_), Out) :-
    t_op(deref, Out).
asm(guard(I), Out) :- 
    t_op(guard, Out),
    emit('  :L~d\n', [I], Out).
asm(idle(I), Out) :- 
    t_op(idle, Out),
    clear_a,
    emit('  :L~d\n', [I], Out).
asm(if_false(I), Out) :- 
    t_op(isfalse, Out),
    emit('  :L~d\n', [I], Out).
asm(if_true(I), Out) :- 
    t_op(istrue, Out),
    emit('  :L~d\n', [I], Out).
asm(int_lit(N), Out) :- 
    clear_a, 
    t_op(lit, Out),
    hex(N, X),
    emit('  ~a ( int )\n', [X], Out).
asm(intval, Out) :- 
    clear_a, 
    t_op(intval, Out).
asm(intval_checked, Out) :- 
    clear_a, 
    t_op(intval_checked, Out).
asm(invoke_goal, Out) :- t_op(invoke_goal, Out).
asm(invoke_jump, Out) :- t_op(invoke_jump, Out).
asm(isatom(A, L, LF, _), Out) :- 
    resolve_atom_index(A, At),
    t_op(isval, Out),
    emit('  ~a :L~d :L~d\n', [At, L, LF], Out).
asm(isint(N, L, LF), Out) :-
    N >= -16384, N =< 16383,
    !,
    X is (N << 1) \/ 0x8001,
    t_op(isval, Out),
    hex(X, XX),
    emit('  ~a ( ~d )\n  :L~d :L~d\n', [XX, N, L, LF], Out).
asm(isint(N, L, LF), Out) :-
    X is ((N << 1) /\ 0x7fff) \/ 0x8001,
    t_op(isint, Out),
    hex(X, XX),
    emit('  ~a :L~d :L~d\n', [XX, L, LF], Out).
asm(islist(L, LF), Out) :- 
    t_op(islist, Out),
    clear_a,
    emit('  :L~d :L~d\n', [L, LF], Out).    
asm(islist_n(L), Out) :- 
    t_op(islist_n, Out),
    emit('  :L~d\n', [L], Out).
asm(islist_t(L), Out) :- 
    t_op(islist_t, Out),
    emit('  :L~d\n', [L], Out).
asm(istuple(N, L, LF), Out) :-
    t_op(istuple, Out),
    clear_a,
    hex(N, X),
    emit('  ~a :L~d :L~d\n', [X, L, LF], Out).
asm(jump(L), Out) :- 
    t_op(jump, Out),
    emit('  :L~d\n', [L], Out).
asm(label(L), Out) :- 
    clear_a,
    emit('@L~d\n', [L], Out).
asm(literal(L), Out) :- 
    clear_a,
    t_op(lit, Out),
    emit('  :C~d\n', [L], Out).
asm(lookup(N, Args, Ls, Lo), Out) :-
    t_op(lookup, Out),
    hex(N, X),
    emit('  ~a\n', [X], Out),
    emit_lookup_table(Args, Ls, Out),
    emit('  :L~d\n', [Lo], Out).
asm(match, Out) :- 
    clear_a,
    t_op(match, Out).
asm(mkint(I), Out) :- 
    I >= -16384, I =< 16383,
    !,
    X is ((I << 1) \/ 0x8001) /\ 0xffff,
    clear_a,
    hex(X, XX),
    t_op(lit, Out),
    emit('  ~a ( ~d )\n', [XX, I], Out).
asm(mkint(I), _) :- 
    error('integer literal exceeds representable range: ~d\n', [I]).
asm(mkint_val(I), Out) :- 
    clear_a,
    hex(I, XX),
    t_op(lit, Out),
    emit('  ~a ( val: ~d )\n', [XX, I], Out).
asm(mklist, Out) :- 
    clear_a,
    t_op(mklist, Out).
asm(mktuple(I, Arity, _), Out) :- 
    clear_a,
    resolve_atom_index(I, A),
    t_op(mktuple, Out),
    hex(Arity, X),
    emit('  ~a ~a\n', [A, X], Out).
asm(mktuple(Arity), Out) :- 
    clear_a,
    t_op(mktuple0, Out),
    hex(Arity, X),
    emit('  ~a\n', [X], Out).
asm(mkvar, Out) :- 
    clear_a,
    t_op(mkvar, Out).
asm(ngoal, _).
asm(not_suspending(L), Out) :-
    t_op(not_suspending, Out),
    emit('  :L~d\n', [L], Out).
asm(open_goal(I, N, A), Out) :-
    t_op(opengoal, Out),
    clear_a,
    emit('  :P~d ( call: ~a/~d )\n', [I, N, A], Out).
asm(pop, Out) :- 
    clear_a,
    t_op(pop, Out).
asm(push, Out) :- 
    t_op(push, Out).
asm(putarg(I), Out) :- 
    I2 is I + 1,
    format_to_atom(Op, 'putarg~d', [I2]),
    t_op(Op, Out).
asm(putarg_bind(I, J), Out) :-
    set_a(e(J)),
    t_op(move_var_e_a, Out),
    JI is ((J * 2) << 8) \/ ((I + 1) * 2),
    hex(JI, X),
    emit('  ~a ( mkvar -> ENV~d -> ARG~d )\n', [X, J, I], Out).
asm(putarg_env(I, J), Out) :-
    t_op(move_e_a, Out),
    JI is ((J * 2) << 8) \/ ((I + 1) * 2),
    hex(JI, X),
    set_a(e(J)),
    emit('  ~a ( ENV~d -> ARG~d )\n', [X, J, I], Out).
asm(putenv_arg(I, J), Out) :- 
    t_op(move_a_e, Out),
    IJ is (((I + 1) * 2) << 8) \/ (J * 2),
    hex(IJ, X),
    emit('  ~a ( ARG~d -> ENV~d )\n', [X, I, J], Out),
    set_a(e(J)).
asm(putenvu(I), Out) :- 
    resolve_addr('ENV', A),
    set_a(e(I)),
    t_op(poke, Out),
    N is A + I * 2,
    hex(N, X),
    emit('  ~a ( ENV~d )\n', [X, I], Out).
asm(putparam(I), Out) :- 
    I2 is I + 1,
    format_to_atom(Op, 'putparam~d', [I2]),
    t_op(Op, Out).
asm(putparam_bind(I, J), Out) :-
    t_op(move_var_e_p, Out),
    JI is ((J * 2) << 8) \/ (I * 2),
    hex(JI, X),
    set_a(e(J)),
    emit('  ~a ( mkvar -> ENV~d -> PARAM~d )\n', [X, J, I], Out).
asm(putparam_env(I, J), Out) :-
    t_op(move_e_p, Out),
    JI is ((J * 2) << 8) \/ (I * 2),
    hex(JI, X),
    set_a(e(J)),
    emit('  ~a ( ENV~d -> PARAM~d )\n', [X, J, I], Out).
asm(puttailarg, Out) :-
    t_op(puttailarg, Out).
asm(reset, Out) :- t_op(reset, Out).
asm(storearg(I), Out) :- 
    I2 is I + 1,
    format_to_atom(Op, 'storearg~d', [I2]),
    t_op(Op, Out).
asm(suspend, Out) :- t_op(suspend, Out).
asm(suspending(L), Out) :-
    t_op(suspending, Out),
    emit('  :L~d\n', [L], Out).
asm(switch_atomic(L), Out) :- 
    t_op(isatomic, Out),
    emit('  :L~d\n', [L], Out).
asm(switch_list(L), Out) :- 
    t_op(islist_t, Out),
    emit('  :L~d\n', [L], Out).
asm(switch_tuple(L), Out) :- 
    t_op(istuple_t, Out),
    emit('  :L~d\n', [L], Out).
asm(tailgoal(I, N, A), Out) :-
    t_op(tailgoal, Out), 
    emit('  :P~d ( tailcall: ~a/~d )\n', [I, N, A], Out).
asm(tailjump, Out) :- t_op(tailjump, Out).
asm(test_number, Out) :- 
    clear_a,
    t_op(test_number, Out).
asm(test_atom, Out) :- 
    clear_a,
    t_op(test_atom, Out).
asm(test_list, Out) :- 
    clear_a,
    t_op(test_list, Out).
asm(test_tuple, Out) :- 
    clear_a, 
    t_op(test_tuple, Out).
asm(test_var, Out) :- 
    clear_a,
    t_op(test_var, Out).
asm(to_int, Out) :- 
    clear_a,
    t_op(to_int, Out).
asm(try_else(L), Out) :- 
    clear_a,
    emit('@L~d\n', [L], Out),
    t_op(tryelse, Out).
asm(unify, Out) :- 
    clear_a,
    t_op(unify, Out).
asm(unify_checked, Out) :- 
    clear_a,
    t_op(unify_checked, Out).
asm(unref(I), Out) :- 
    resolve_addr('PARAM0', A),
    t_op(drop, Out),
    N is A + I * 2,
    hex(N, X),
    emit('  ~a ( PARAM~d )\n', [X, I], Out).
asm(yield, Out) :- t_op(yield, Out).
asm(x_assign(I), Out) :- 
    resolve_addr(fl_xassign, A),
    N is I * 2,
    hex(A, X),
    hex(N, NX),
    emit('  #~a ( #ENV~d )  #~a JSR2 ( fl_xassign )\n', [NX, I, X], Out).
asm(x_getregister(_), _).
asm(x_putasmparam(I, _), Out) :- 
    resolve_addr('ENV', A),
    E is A + I * 2,
    hex2(E, EX),
    emit('  #~a LDZ2 ( ENV~d )\n', [EX, I], Out),
    xcall(fl_deref, Out).

t_op(Opname, Out) :- 
    format_to_atom(L, 'op_~a', [Opname]),
    resolve_addr(L, Addr),
    hex(Addr, X),
    emit('  ~a ( op_~a )\n', [X, Opname], Out).

t_op(Opname, I, Out) :- 
    format_to_atom(L, 'op_~a~d', [Opname, I]),
    resolve_addr(L, Addr),
    hex(Addr, X),
    emit('  ~a ( op_~a~d )\n', [X, Opname, I], Out).

xcall(Name, Out) :-
    resolve_addr(Name, Addr),
    hex(Addr, X),
    emit('  #~a JSR2 ( ~a )\n', [X, Name], Out).

assign_labels([], L, L).
assign_labels([L1|R], L1, L) :-
    L2 is L1 + 1,
    assign_labels(R, L2, L).

dump_resolved_atoms(_).

% preallocated literal dumping
dump_literal([X|Y], L, Out) :-
    !,
    emit('  * @C~d 20ff\n', [L], Out),
    dump_sub_literal(X, Out),
    dump_sub_literal(Y, Out).
dump_literal(X, L, Out) :-
    integer(X),
    !,
    hex(X, XX),
    emit('  * @C~d 10ff ~a 0000\n', [L, XX], Out).
dump_literal(functor(N, A, AL), L, Out) :-
    Tag is 0x40ff \/ (A << 8),
    hex(Tag, Tx),
    emit('  * @C~d ~a\n', [L, Tx], Out),
    emit('  :S~d :C~d\n', [N, AL], Out).
dump_literal(tuple(A, HL, AL), L, Out) :-
    Tag is 0x40ff \/ (A << 8),
    hex(Tag, Tx),
    emit('  * @C~d ~a\n', [L, Tx], Out),
    dump_sub_literal(HL, Out),
    emit('  :C~d\n', [AL], Out).

atom_label([], L) :- !, resolve_addr(fl_nil, L1), hex(L1, L).
atom_label('[]', L) :- !, resolve_addr(fl_nil, L1), hex(L1, L).
atom_label(true, L) :- !, resolve_addr(fl_true, L1), hex(L1, L).
atom_label(false, L) :- !, resolve_addr(fl_false, L1), hex(L1, L).
atom_label(S, L) :-
    r_atom(S, A),
    format_to_atom(L, ':S~d', [A]).

resolve_atom_index(0, A) :- !, resolve_addr(fl_nil, A1), hex(A1, A).
resolve_atom_index(1, A) :- !, resolve_addr(fl_true, A1), hex(A1, A).
resolve_atom_index(2, A) :- !, resolve_addr(fl_false, A1), hex(A1, A).
resolve_atom_index(N, A) :- format_to_atom(A, ':S~d', [N]).

dump_sub_literal(X, Out) :-
    integer(X), X >= -16384, X =< 16383,
    !,
    N is (X << 1) \/ 0x8001,
    hex(N, NX),
    emit('  ~a ( ~d )\n', [NX, X], Out).
dump_sub_literal([], Out) :-
    % for swipl
    !,
    atom_label('[]', A),
    emit('  ~a\n', [A], Out).
dump_sub_literal(X, Out) :-
    atom(X), 
    !,
    atom_label(X, A),
    emit('  ~a\n', [A], Out).
dump_sub_literal(label(L), Out) :-
    emit('  :C~d\n', [L], Out).

emit_list([], _).
emit_list([X], Out) :- 
    hex2(X, XX),
    emit('~a ', [XX], Out).
emit_list([X|L], Out) :- 
    hex2(X, XX),
    emit('~a ', [XX], Out),
    emit_list(L, Out).

resolve_addr(Name, Addr) :- rt_addr(Name, Addr), !.
resolve_addr(Name, _) :-
    error('unresolved address: ~a\n', [Name]).

emit_entry_point(Out) :-
    bagof(Name, module_init(Name), Names),
    emit_entry_points(Names, Out).
emit_entry_point(_) :-
    error('initialization goal required\n', []).

emit_entry_points([], Out) :- 
    resolve_addr('op_yield', Addr),
    hex(Addr, X),
    emit('  #~a JMP2 ( yield )\n', [X], Out).
emit_entry_points([Name|L], Out) :-
    pdef(Name/0, I),
    !,
    resolve_addr(fl_initgoal, A),
    hex(A, Ax),
    emit('  ;P~d #~a JSR2 ( fl_initgoal )\n', [I, Ax], Out),
    emit_entry_points(L, Out).
emit_entry_points([Name|_], _) :-
    error('initialization goal ~a/0 is not defined\n', [Name]).

emit_foreign_files(Out) :-
    foreign_file(Name),
    emit('~~~a\n', [Name], Out),
    fail.
emit_foreign_files(_).

emit_lookup_table([], _, _).
emit_lookup_table([Arg|Args], [L|Ls], Out) :-
    atom(Arg),
    register_atom(Arg, Lit),
    resolve_atom_index(Lit, A),
    emit('  ~a :L~d\n', [A, L], Out),
    emit_lookup_table(Args, Ls, Out).
emit_lookup_table([Arg|Args], [L|Ls], Out) :-
    X is ((Arg << 1) \/ 0x8001) /\ 0xffff,
    hex(X, XX),
    emit('  ~a :L~d\n', [XX, L], Out),
    emit_lookup_table(Args, Ls, Out).

hex(N, X) :-
    H is (N >> 8) /\ 0xff,
    L is N /\ 0xff,
    hex2(H, Hx),
    hex2(L, Lx),
    atom_concat(Hx, Lx, X).

hex2(N, X) :-
    N > 15,
    !,
    format_to_atom(X, '~16r', [N]).
hex2(N, X) :-
    format_to_atom(X, '0~16r', [N]).

:- include('rt0uxn.pl').
